from __future__ import annotations
import typing
import struct
from dataclasses import dataclass
from typing import Callable
from bytemaker.utils import is_subclass_of_union
from bytemaker.bits import Bits


class PyTypeMeta(type):
    """
    This is used to create IsByteConvertible, a type to allow checking \
        whether an object or instances of a class can be converted to a bytes object using isinstance or issubclass.
    """
    def __instancecheck__(self, __instance) -> bool:
        return ConversionConfig.has_suitable_conversion(type(__instance))

    def __subclasscheck__(self, __subclass: type) -> bool:
        return ConversionConfig.has_suitable_conversion(__subclass)


class PyType(metaclass=PyTypeMeta):
    pass


# PyType Handling
@dataclass
class ConversionInfo:
    pytype: type
    to_bits: Callable[[typing.Any], Bits]
    from_bits: Callable[[Bits], typing.Any]
    num_bits: Callable[[typing.Any], int]

    @classmethod
    def num_bytes(cls, typeinstance) -> int:
        default = (cls.num_bits(typeinstance) + 7) // 8
        return default if default > 0 else 1

    @classmethod
    def to_bytes(cls, pytype) -> bytes:
        return bytes(cls.to_bits(pytype))

    @classmethod
    def from_bytes(cls, bytes_obj) -> typing.Any:
        return cls.from_bits(Bits(bytes_obj))


class ConversionConfig:
    """
    Class to configure conversions for Python primitives.
    """
    _implemented_conversions: dict[type, ConversionInfo] = {}
    _known_furthest_descendant_mappings: dict[type, type] = {}
    _has_a_suitable_conversion: dict[type, bool] = {}

    @classmethod
    def set_conversion_info(cls, conversion_info: ConversionInfo):
        # # If the conversion info pytype is an exact match for an already-mapped type,
        # #   replace any prior mappings to superclasses for that pytype with the new conversion
        # if conversion_info.pytype in cls._implemented_conversions:
        #     cls._known_furthest_descendant_mappings[conversion_info.pytype] = conversion_info.pytype

        # If the conversion info pytype is a stricter subclass of an already-mapped type,
        #   replace the mapping for the superclass with the new conversion
        for key, value in cls._known_furthest_descendant_mappings.items():
            could_map_key_to_conv_pytype_conversion = is_subclass_of_union(key, conversion_info.pytype)
            conv_pytype_is_stricter_match_than_existing = is_subclass_of_union(conversion_info.pytype, value)
            if could_map_key_to_conv_pytype_conversion and conv_pytype_is_stricter_match_than_existing:
                cls._known_furthest_descendant_mappings[key] = conversion_info.pytype

        # Set the conversion info
        cls._implemented_conversions[conversion_info.pytype] = conversion_info
        cls._known_furthest_descendant_mappings[conversion_info.pytype] = conversion_info.pytype
        cls._has_a_suitable_conversion[conversion_info.pytype] = True

        # Check types previously ascertained to have no suitable conversion
        #  if this new version involves a superclass of that type, set this new conversion type
        #  as the furthest descendant mapping for that type and flag that type as a suitable conversion
        types_known_to_not_have_suitable_conversion = [
            pytype for pytype, has_suitable_conversion in cls._has_a_suitable_conversion.items()
            if not has_suitable_conversion]

        for pytype in types_known_to_not_have_suitable_conversion:
            if is_subclass_of_union(conversion_info.pytype, pytype):
                cls._known_furthest_descendant_mappings[pytype] = conversion_info.pytype
                cls._has_a_suitable_conversion[pytype] = True

    @classmethod
    def has_suitable_conversion(cls, pytype: type) -> bool:
        if pytype in cls._has_a_suitable_conversion:
            return cls._has_a_suitable_conversion[pytype]
        else:
            for implemented_pytype in cls._implemented_conversions.keys():
                if is_subclass_of_union(pytype, implemented_pytype):
                    cls._has_a_suitable_conversion[pytype] = True
                    return True

    @classmethod
    def get_conversion_info(cls, pytype: type) -> ConversionInfo:
        # If the pytype is an exact match for a conversion, return that conversion
        if pytype in cls._implemented_conversions:
            return cls._implemented_conversions[pytype]

        # If the pytype is a known subclass of a conversion, return the conversion for the superclass
        if pytype in cls._known_furthest_descendant_mappings:
            return cls._implemented_conversions[cls._known_furthest_descendant_mappings[pytype]]

        # If the pytype is a subclass of a conversion, return the conversion for the superclass
        if cls.has_suitable_conversion(pytype):
            cur_suitable_implemented_pytype = None
            for candidate_implemented_pytype in cls._implemented_conversions.keys():
                pytype_is_subclass_of_candidate = is_subclass_of_union(pytype, candidate_implemented_pytype)
                candidate_is_stricter_than_current = (
                    cur_suitable_implemented_pytype is None or
                    is_subclass_of_union(candidate_implemented_pytype, cur_suitable_implemented_pytype)
                )
                if pytype_is_subclass_of_candidate and candidate_is_stricter_than_current:
                    cur_suitable_implemented_pytype = candidate_implemented_pytype

            if cur_suitable_implemented_pytype is not None:
                cls._known_furthest_descendant_mappings[pytype] = cur_suitable_implemented_pytype

            return cls._implemented_conversions[cls._known_furthest_descendant_mappings[pytype]]
        return None


# _string_conversion_info = ConversionInfo(
#     pytype=str,
#     to_bits=lambda string: Bits(string.encode('utf-8')),
#     from_bits=lambda bits: bits.to_bytes().decode('utf-8'),
#     num_bits=lambda string: len(string.encode('utf-8')) * 8
# )
# ConversionConfig.set_conversion_info(_string_conversion_info)

_char_conversion_info = ConversionInfo(
    pytype=str,
    to_bits=lambda string: Bits(string.encode('utf-8')),
    from_bits=lambda bits: bits.to_bytes().decode('utf-8'),
    num_bits=8
)
ConversionConfig.set_conversion_info(_char_conversion_info)

for bytesish in [bytes, bytearray, memoryview]:
    conversion_info = ConversionInfo(
        pytype=bytesish,
        to_bits=lambda bys: Bits(bys),
        from_bits=lambda bits: bits.to_bytes(),
        num_bits=lambda bys: len(bys) * 8
    )
    # ConversionConfig.set_conversion_info(conversion_info)

bool_conversion_info = ConversionInfo(
    pytype=bool,
    to_bits=lambda boo: Bits([int(boo)]),
    from_bits=lambda bits: bool(bits.to_int()),
    num_bits=1
)
ConversionConfig.set_conversion_info(bool_conversion_info)


# def _int_to_bits(num: int) -> Bits:
#     if issubclass(type(num), bool):
#         to_convert = int(num)
#     return Bits(to_convert, to_convert.to_bytes(twos_complement_bit_length(num), byteorder='little', signed=num >= 0))


int_conversion_info = ConversionInfo(
    pytype=int,
    to_bits=lambda num: Bits.from_int(num, size=32),
    from_bits=lambda bits: bits.to_int(),
    num_bits=32
)
ConversionConfig.set_conversion_info(int_conversion_info)


float_conversion_info = ConversionInfo(
    pytype=float,
    to_bits=lambda fl: Bits(struct.pack('>f', fl)),
    from_bits=lambda bits: struct.unpack('>f', bits.to_bytes())[0],
    num_bits=32
)
ConversionConfig.set_conversion_info(float_conversion_info)


def pytype_to_bits(py_prim: type) -> Bits:
    """
    Function to convert Python instances into a default number of Bits.
        Uses the conversions in ConversionConfig.

    Args:
        py_prim: The python instance to convert to Bits

    Returns:
        Bits: The Bits representation of the python instance
    """
    py_prim_type = type(py_prim)

    conversion = ConversionConfig.get_conversion_info(py_prim_type)

    if conversion is None:
        raise TypeError(f"No conversion found for {py_prim_type}")

    return conversion.to_bits(py_prim)


def pytype_to_bytes(py_prim: type, reverse_endianness: bool = False) -> bytes:
    """
    Function to convert Python instances into a default number of bytes.
        Uses the conversions in ConversionConfig.

    Args:
        py_prim: The python instance to convert to bytes

    Returns:
        bytes: The bytes representation of the python instance
    """
    retval = pytype_to_bits(py_prim).to_bytes()
    if reverse_endianness:
        retval = retval[::-1]
    return retval


def bits_to_pytype(bits_obj: Bits, pytype: type):
    """
    Function to convert bits into instances of Python types.

    Args:
        bytes_obj (bytes): The bits object to convert to a Python primitive
        py_prim_type (type): The type of the Python primitive to convert to. Must be a member of PyTypeWithDefaultBytes

    Returns:
        pytype: The instance of thee provided Python type represented by the bits
    """

    conversion = ConversionConfig.get_conversion_info(pytype)

    if conversion is None:
        raise TypeError(f"No conversion found for {pytype}")

    return conversion.from_bits(bits_obj)


def bytes_to_pytype(bytes_obj: Bits, pytype: type):
    """
    Function to convert bytes into instances of Python types.

    Args:
        bytes_obj (bytes): The bytes object to convert to a Python primitive
        py_prim_type (type): The type of the Python primitive to convert to. Must be a member of PyTypeWithDefaultBytes

    Returns:
        pytype: The instance of thee provided Python type represented by the bytes
    """

    return bits_to_pytype(Bits(bytes_obj), pytype)



# CType Handling
from ctypes import _SimpleCData, Structure, Union, Array
import ctypes
import typing
from bytemaker.utils import is_instance_of_union, is_subclass_of_union
from bytemaker.bits import Bits


CType = typing.Union[_SimpleCData, Structure, Union, Array]


def reverse_bytes_unit(unit: _SimpleCData):
    """ Reverse the byte order of the given value. """
    unit_type = type(unit)
    bytes_value = bytearray(unit)
    bytes_value.reverse()
    return unit_type.from_buffer_copy(bytes_value)


def reverse_ctype_endianness(ctype_instance: CType) -> None:
    """ Process each field of the structure. """

    if isinstance(ctype_instance, _SimpleCData):
        # Reverse the byte order for single, multi-byte objects
        byte_size = ctypes.sizeof(ctype_instance)
        if byte_size > 1:
            ctype_instance = reverse_bytes_unit(ctype_instance)

    if isinstance(ctype_instance, Array):
        for i in range(len(ctype_instance)):
            ctype_instance[i] = reverse_ctype_endianness(ctype_instance[i])

    if isinstance(ctype_instance, Structure):
        for field_name, field_type in ctype_instance._fields_:
            field_value = getattr(ctype_instance, field_name)
            # print(field_name, field_value, type(field_value))
            reversed_unit = reverse_ctype_endianness(field_type(field_value))
            setattr(ctype_instance, field_name, reversed_unit)

    return ctype_instance


def ctype_to_bytes(ctype_obj: CType, reverse_endianness=True) -> bytes:
    """
    Function to convert ctypes into bytes objects

    Args:
        ctype_obj (ctypes._SimpleCData | ctypes.Structure | ctypes.Union | ctypes.Array):
            The ctypes object to convert to bytes
        reverse_endianness (bool, optional):
            Whether to reverse the endianness of the bytes after converting. Defaults to False.

    Returns:
        bytes: The bytes representation of the ctypes object
    """
    if not is_instance_of_union(ctype_obj, CType):
        raise TypeError(
            f"ctype_to_bytes only accepts _SimpleCData, Structure, Union, and Array objects, not {type(ctype_obj)}"
        )

    if reverse_endianness:
        ctype_obj = reverse_ctype_endianness(ctype_obj)

    retbytes = bytes(ctype_obj)

    return retbytes


def ctype_to_bits(ctype_obj: CType, reverse_endianness=True) -> Bits:
    """
    Function to convert ctypes into Bits objects

    Args:
        ctype_obj (ctypes._SimpleCData | ctypes.Structure | ctypes.Union | ctypes.Array):
            The ctypes object to convert to Bits
        reverse_endianness (bool, optional):
            Whether to reverse the endianness of the Bits after converting. Defaults to False.

    Returns:
        Bits: The Bits representation of the ctypes object
    """
    return Bits(ctype_to_bytes(ctype_obj, reverse_endianness=reverse_endianness))


def bytes_to_ctype(bytes_obj: bytes,
                   ctype_type: type,
                   reverse_endianness=True) -> CType:
    """
    Function to convert bytes into ctypes objects

    Args:
        bytes_obj (bytes): The bytes object to convert to a ctypes object
        ctype_type (type): The type of the ctypes object to convert to.
            Must be a member of CType
        reverse_endianness (bool, optional): Whether to reverse the endianness
            of the bytes before converting. Defaults to False.

    Returns:
        ctypes._SimpleCData | ctypes.Structure | ctypes.Union | ctypes.Array:
            The ctypes object representation of the bytes
    """

    if not is_subclass_of_union(ctype_type, CType):
        raise TypeError(
            f"bytes_to_ctype only accepts _SimpleCData, Structure, Union, and Array types, not {ctype_type}"
        )

    ctype_obj = ctype_type.from_buffer_copy(bytes_obj)
    if reverse_endianness:
        ctype_obj = reverse_ctype_endianness(ctype_obj)

    return ctype_obj


def bits_to_ctype(
        bits_obj: Bits,
        ctype_type: type,
        reverse_endianness=True) -> CType:
    """
    Function to convert bits into ctypes objects

    Args:
        bits_obj (Bits): The bits object to convert to a ctypes object
        ctype_type (type): The type of the ctypes object to convert to.
            Must be a member of CType
        reverse_endianness (bool, optional): Whether to reverse the endianness
            of the bits before converting. Defaults to False.

    Returns:
        ctypes._SimpleCData | ctypes.Structure | ctypes.Union | ctypes.Array:
            The ctypes object representation of the bits
    """
    return bytes_to_ctype(bits_obj.to_bytes(), ctype_type, reverse_endianness=reverse_endianness)





import ctypes
import dataclasses
from typing import Iterable, Union
from bytemaker.bits import Bits
from bytemaker.utils import is_instance_of_union, is_subclass_of_union, DataClassType
from bytemaker.ytypes import YType, bytes_to_ytype
from bytemaker.native_types.ctypes_ import CType, bytes_to_ctype, ctype_to_bytes, bits_to_ctype, ctype_to_bits
from bytemaker.native_types.pytypes import (
    PyType, pytype_to_bits, pytype_to_bytes, bits_to_pytype, bytes_to_pytype, ConversionConfig
)


UnitType = Union[CType, YType, PyType]

# CType is a Union of _SimpleCData, Structure, Union, and Array
# YType is a Union of YInt, YFloat, YString, YBytes, YBool, YEnum, YArray, and YStruct
# PyType is a Union of int, float, str, bytes, bool, and Enum


def count_bits_in_unit_type(unit_type: UnitType) -> int:
    """
    Function to count the number of bits in a UnitType- a Python, type, ctype, or YType (bytemaker type).
    """

    # print("Counting bits in unit type", unit_type)
    if is_subclass_of_union(unit_type, CType):
        return ctypes.sizeof(unit_type) * 8
    elif is_subclass_of_union(unit_type, YType):
        return unit_type.num_bits
    elif is_subclass_of_union(unit_type, PyType):
        # print(ConversionConfig.get_conversion_info(unit_type).num_bits)
        return ConversionConfig.get_conversion_info(unit_type).num_bits
    elif is_subclass_of_union(unit_type, DataClassType):
        size_in_bits = 0
        for field in dataclasses.fields(unit_type):
            field_type = field.type
            if isinstance(field_type, str):
                field_type = eval(field_type)
            size_in_bits += count_bits_in_unit_type(field_type)
        return size_in_bits


def count_bits_in_aggregate_type(aggregate_type: type) -> int:
    """
    Function to count the number of bits in an aggregate type- a Python, type, ctype, or YType (bytemaker type).
    """
    if is_subclass_of_union(aggregate_type, UnitType):
        return count_bits_in_unit_type(aggregate_type)
    else:
        size_in_bits = 0
        for field in dataclasses.fields(aggregate_type):
            field_type = field.type
            if isinstance(field_type, str):
                field_type = eval(field_type)
            size_in_bits += count_bits_in_unit_type(field_type)
        return size_in_bits


def count_bytes_in_unit_type(unit_type: UnitType) -> int:
    """
    Function to count the number of bytes in a UnitType-
        a Python numeric/binary/string type, ctype, or YType (bytemaker type).
    """
    return (count_bits_in_unit_type(unit_type) + 1) // 8


def to_bits_individual(unit: UnitType) -> Bits:
    """
    Function to convert a single Python primitive or ctypes object into Bits.
    """
    if is_instance_of_union(unit, CType):
        return ctype_to_bits(unit)
    elif isinstance(unit, YType):
        return unit.to_bits()
    elif is_instance_of_union(unit, PyType):
        return pytype_to_bits(unit)
    else:
        raise Exception(f"Cannot convert {unit} to bits because the unit type is not a CType, YType, or PyType")


def to_bytes_individual(unit: UnitType, reverse_endianness: bool = False) -> bytes:
    """
    Function to convert a single Python primitive or ctypes object into bytes.
    """

    if is_instance_of_union(unit, CType):
        return ctype_to_bytes(unit, reverse_endianness=reverse_endianness)
    elif isinstance(unit, YType):
        return unit.to_bytes(reverse_endianness=reverse_endianness)
    elif is_instance_of_union(unit, PyType):
        return pytype_to_bytes(unit, reverse_endianness=reverse_endianness)
    else:
        raise Exception(f"Cannot convert {unit} to bytes because the unit type is not a CType, YType, or PyType")


def from_bits_individual(unitbits: Bits, unittype: type) -> PyType:
    """
    Function to convert Bits into a single UnitType-
        a Python numeric/binary/string type, ctype, or YType (bytemaker type).

    Args:
        unitbits (Bits): The Bits object to convert to a UnitType
        unittype (type): The type of the UnitType to convert to. Must be a member of UnitType
    """

    size_in_bits = count_bits_in_unit_type(unittype)

    if len(unitbits) != size_in_bits:
        raise Exception(f"Cannot convert {unitbits} to {unittype}"
                        f" because the number of bits in the bits object ({unitbits.num_bits})"
                        f" does not match the number of bits in the unit type ({size_in_bits})")
    if is_subclass_of_union(unittype, CType):
        return bits_to_ctype(unitbits, unittype)
    elif is_subclass_of_union(unittype, YType):
        return unittype.from_bits(unitbits)
    elif is_subclass_of_union(unittype, PyType):
        return bits_to_pytype(unitbits, unittype)
    else:
        raise Exception(f"Cannot convert {unitbits} to {unittype}"
                        f" because the unit type is not a CType, YType, or PyType")


def from_bytes_individual(unitbytes: bytes, unittype: type, reverse_endianness: bool = False) -> PyType:
    """
    Function to convert bytes into a single UnitType-
        a Python numeric/binary/string type, ctype, or YType (bytemaker type).

    Args:
        unitbytes (bytes): The bytes object to convert to a UnitType
        unittype (type): The type of the UnitType to convert to.
            Must be a member of UnitType
        reverse_endianness (bool, optional): Whether to reverse the endianness of the bytes before converting.
            Defaults to False.
    """

    size_in_bits = count_bits_in_unit_type(unittype)
    if len(unitbytes) * 8 != size_in_bits:
        raise Exception(f"Cannot convert {unitbytes} to {unittype}"
                        f" because the number of bits in the bytes object ({len(unitbytes) * 8})"
                        f" does not match the number of bits in the unit type ({size_in_bits})")
    if is_subclass_of_union(unittype, CType):
        return bytes_to_ctype(unitbytes, unittype, reverse_endianness=reverse_endianness)
    elif is_subclass_of_union(unittype, YType):
        return bytes_to_ytype(unitbytes, unittype, reverse_endianness=reverse_endianness)
    elif is_subclass_of_union(unittype, PyType):
        return bytes_to_pytype(unitbytes, unittype, reverse_endianness=reverse_endianness)
    else:
        raise Exception(f"Cannot convert {unitbytes} to {unittype}"
                        f" because the unit type is not a CType, YType, or PyType")


AggregateTypeByteConvertible = Union[DataClassType, YType, CType, PyType, Iterable]


def trycast(obj, type_):
    try:
        return type_(obj)
    except TypeError:
        return obj


def to_bits_aggregate(convertible_object: AggregateTypeByteConvertible) -> Bits:
    """
    Function to convert fixed-byte Python types (int, float, char), ctypes, and YTypes into Bits.
        Also converts dataclasses annotated with these types into bits (this support is recursive).
        Technically supports Iterables as well, but note that from_bits_aggregate will not work on Iterables.

    Args:
        units (DataClassType | YType | CType | PyType | Iterable): The object to convert to Bits

    Returns:
        Bits: The Bits representation of the object
    """

    ret_bits = Bits()

    # print("to_bits_aggregate", convertible_object)
    # print("type(units)", type(convertible_object))
    # print("isinstance(units, DataClassType)", isinstance(convertible_object, DataClassType))

    if (
        is_instance_of_union(convertible_object, UnitType) and not
        (isinstance(convertible_object, str) and len(convertible_object) > 1)
    ):
        ret_bits = to_bits_individual(convertible_object)
    elif isinstance(convertible_object, DataClassType):
        fields = dataclasses.fields(convertible_object)
        field_values = [getattr(convertible_object, field.name) for field in fields]
        field_types = [field.type if not isinstance(field.type, str) else eval(field.type) for field in fields]
        # print("types", field_types)
        # print("type_is_dataclass", [isinstance(field_type, DataClassType) for field_type in field_types])
        field_values = [trycast(field_value, field_type) for field_type, field_value in zip(field_types, field_values)]
        field_value_bits = [to_bits_aggregate(field_value) for field_value in field_values]
        ret_bits = Bits().join(field_value_bits)
    elif isinstance(convertible_object, Iterable):
        for unit in convertible_object:
            ret_bits.extend(to_bits_aggregate(unit))
    else:
        raise Exception(f"Cannot convert {convertible_object} to bits, as it is not a CType, YType, or PyType;"
                        f" or a Dataclass or Iterable of such types. It is a {type(convertible_object)}.")

    return ret_bits


def from_bits_aggregate(unitbits: Bits, aggregate_type: type) -> Union[UnitType, AggregateTypeByteConvertible]:
    if is_subclass_of_union(aggregate_type, UnitType):
        return from_bits_individual(unitbits, aggregate_type)
    else:
        size_in_bits = count_bits_in_aggregate_type(aggregate_type)
        # print("unitbits type", type(unitbits))
        # print(unitbits)
        if len(unitbits) != size_in_bits:
            raise Exception(f"Cannot convert {unitbits} to {aggregate_type}"
                            f" because the number of bits in the bits object ({unitbits.num_bits})"
                            f" does not match the number of bits in the unit type ({size_in_bits})")

        read_fields = list()
        for field in dataclasses.fields(aggregate_type):
            field_type = field.type
            if isinstance(field.type, str):
                field_type = eval(field.type)

            field_size_in_bits = count_bits_in_unit_type(field_type)
            field_bits = unitbits[:field_size_in_bits]
            field_value = from_bits_aggregate(field_bits, field_type)
            read_fields.append(field_value)
            unitbits = unitbits[field_size_in_bits:]
        retval = aggregate_type(*read_fields)

    return retval


def to_bytes_aggregate(units: AggregateTypeByteConvertible, reverse_endianness: bool = False) -> bytes:
    """
    Function to convert a collection of Python primitives or ctypes objects into bytes.

    Args:
        units [Iterable | DataClassType]): The objects to convert to bytes

    Returns:
        bytes: The bytes representation of the objects
    """
    ret_bytes = bytearray()

    if is_instance_of_union(units, UnitType):
        # print("Is unit", type(units))
        ret_bytes = to_bytes_individual(units, reverse_endianness=reverse_endianness)

    elif isinstance(units, DataClassType):
        # print("Is dataclass", type(units))

        for field in dataclasses.fields(units):
            field_type = field.type
            if isinstance(field.type, str):
                field_type = eval(field_type)
            field_value = getattr(units, field.name)
            # print(field_type)
            field_value = field_type(field_value)
            # print("-----")
            try:
                field_value = field_type(field_value)
            except TypeError:
                # print("Couldn't get field value")
                pass
            # print(field)
            # print(field_value)
            # print(field_type)
            # print(type(field_value))
            # print("-----")
            field_value_bytes = to_bytes_aggregate(field_value, reverse_endianness=reverse_endianness)
            # print(field_value_bytes)

            ret_bytes.extend(field_value_bytes)

    elif isinstance(units, Iterable):
        # print("Is iterable")
        # print("Units:", units)
        for unit in units:
            # print(unit)
            ret_bytes.extend(to_bytes_aggregate(unit, reverse_endianness=reverse_endianness))

    return bytes(ret_bytes)


def from_bytes_aggregate(
        bytes_obj: bytes,
        aggregate_type: type,
        is_array=False,
        reverse_endianness: bool = False) -> Union[UnitType, AggregateTypeByteConvertible]:

    if reverse_endianness:
        bytes_obj = bytes_obj[::-1]

    if is_subclass_of_union(aggregate_type, UnitType):
        return from_bytes_individual(bytes_obj, aggregate_type, reverse_endianness=reverse_endianness)
    else:
        size_in_bits = count_bits_in_unit_type(aggregate_type)

        if not is_array:
            if len(bytes_obj) * 8 != size_in_bits:
                raise Exception(f"Cannot convert {bytes_obj} to {aggregate_type}"
                                f" because the number of bits in the bytes object ({len(bytes_obj) * 8})"
                                f" does not match the number of bits in the unit type ({size_in_bits})")

            read_fields = list()
            for field in dataclasses.fields(aggregate_type):
                if isinstance(field.type, str):
                    field_type = eval(field.type)
                field_size_in_bits = count_bits_in_unit_type(field_type)
                field_bytes = bytes_obj[:field_size_in_bits]
                field_value = from_bytes_aggregate(field_bytes, field_type, reverse_endianness=reverse_endianness)
                read_fields.append(field_value)
                bytes_obj = bytes_obj[field_size_in_bits:]
            retval = aggregate_type(*read_fields)

        else:
            arr_entry_list = list()
            for i in range(0, len(bytes_obj), size_in_bits):
                endindex = i + size_in_bits if i + size_in_bits < len(bytes_obj) else len(bytes_obj)
                arr_entry_list.append(from_bytes_aggregate(
                    bytes_obj[i:endindex],
                    aggregate_type,
                    reverse_endianness=reverse_endianness))
            retval = aggregate_type(*arr_entry_list)

    return retval



from __future__ import annotations
from math import ceil
import operator
import typing
from typing import Iterable, Protocol, runtime_checkable
from bytemaker.utils import ByteConvertible, twos_complement_bit_length


@runtime_checkable
class BitsCastable(Protocol):
    def __Bits__(self) -> "Bits":
        pass


BitsConstructorType = typing.Union[int, str, bytes, bytearray, memoryview, Iterable[int], "Bits", BitsCastable]


class Bits:
    """
    A class for storing and manipulating bits.
    """
    def __init__(self, initial_data: BitsConstructorType = None, size=None, deep=False):
        if initial_data is None:
            initial_data = list()

        bitlist: list[int] = None

        if isinstance(initial_data, Bits):
            self.bitlist = initial_data.bitlist

        else:
            if isinstance(initial_data, int):
                bitlist = self.from_int(initial_data, size).bitlist
            elif isinstance(initial_data, BitsCastable):
                if isinstance(initial_data, Bits):
                    bitlist = initial_data.bitlist
                elif hasattr(initial_data, '__Bits__'):
                    bitlist = initial_data.__Bits__().bitlist
            elif isinstance(initial_data, str):
                bitlist = Bits.from_string(initial_data).bitlist
            elif isinstance(initial_data, (bytes, bytearray, memoryview)):
                bitlist = [int(bit) for byte in initial_data for bit in format(byte, '08b')]
            elif isinstance(initial_data, Iterable):  # Iterable[int | bool] in more modern Python versions
                bitlist = [int(bit) for bit in initial_data]
            elif isinstance(initial_data, ByteConvertible):
                bitlist = [int(bit) for byte in bytes(initial_data) for bit in format(byte, '08b')]
            else:
                raise TypeError(f"Cannot convert {type(initial_data)} to Bits. Must be {BitsConstructorType}.")

            self.bitlist = bitlist

            if self.bitlist is None:
                raise ValueError(
                    f"Could not convert {type(initial_data)} to Bits despite the provided type being acceptable. "
                    "The resulting bit_list was none")

            assert all(bit in [0, 1] for bit in self.bitlist), \
                (f"Bits can only be 1 or 0. The provided data produced bits other than 0 or 1."
                 f"Final bitlist: {self.bitlist} Initial data: {initial_data}")

        if deep:
            self.bitlist = self.bitlist.copy()

    @property
    def bytes(self) -> list[int]:
        return bytes(self)

    @property
    def bytearray(self) -> bytearray:
        return bytearray(bytes(self))

    def __str__(self) -> str:
        return self.str_(format_spec='b')

    def str_(self, format_spec) -> str:
        return format(self, format_spec)

    def __repr__(self) -> str:
        return f"Bits({self.bitlist})"

    def __len__(self) -> int:
        return len(self.bitlist)

    def __getitem__(self, index) -> int:
        if isinstance(index, int):
            return self.bitlist[index]
        elif hasattr(index, '__index__'):
            return self.bitlist[index.__index__()]
        elif hasattr(index, '__int__'):
            return self.bitlist[index.__int__()]
        elif isinstance(index, slice):
            return Bits(self.bitlist[index])
        elif isinstance(index, Iterable):
            return Bits([self.bitlist[i] for i in index])
        else:
            raise TypeError(
                f"Index must be an int, slice, or iterable of ints, or have __index__ or __int__ method,"
                f" not {type(index)}"
            )

    def __setitem__(self, index, value):
        if not 0 <= value <= 1:
            raise ValueError('Bits can only be 1 or 0')
        self.bitlist[index] = value

    def binop(self, other: BitsConstructorType, op: "operator", expand=False) -> Bits:
        if not isinstance(other, Bits):
            try:
                other = Bits(other)
            except Exception as e:
                raise TypeError(
                    f"Can only perform bitwise operations on Bits objects, not {type(other)}"
                    f"Exception details {e}"
                )

        bitsleftlen = len(self)
        bitsrightlen = len(other)
        bitsleftiter = iter(self)
        bitsrightiter = iter(other)

        retbits = list()

        while bitsleftlen < bitsrightlen:
            if expand:
                retbits.append(next(bitsrightiter))
            bitsrightlen -= 1

        while bitsrightlen < bitsleftlen:
            if expand:
                retbits.append(next(bitsleftiter))
            bitsleftlen -= 1

        while bitsleftlen > 0:
            retbits.append(op(next(bitsleftiter), next(bitsrightiter)))
            bitsleftlen -= 1

        return Bits(retbits)

    def __invert__(self) -> Bits:
        return Bits([int(not bit) for bit in self.bitlist])

    def and_(self, other: Bits, expand=False) -> Bits:
        return self.binop(other, operator.and_, expand=expand)

    def or_(self, other: Bits, expand=False) -> Bits:
        return self.binop(other, operator.or_, expand=expand)

    def xor_(self, other: Bits, expand=False) -> Bits:
        return self.binop(other, operator.xor, expand=expand)

    def __and__(self, other: Bits) -> Bits:
        return self.and_(other)

    def __or__(self, other: Bits) -> Bits:
        return self.or_(other)

    def __xor__(self, other: Bits) -> Bits:
        return self.xor_(other)

    def lshift(self, other: int, expand=False) -> Bits:
        if expand:
            return Bits(self.bitlist + [0] * other)
        else:
            return Bits(self.bitlist[other:] + [0] * other)

    def __lshift__(self, other: int) -> Bits:
        return self.lshift(other)

    def rshift(self, other: int, expand=False) -> Bits:
        if expand:
            return Bits([0] * other + self.bitlist)
        else:
            return Bits([0] * other + self.bitlist[:-other])

    def __rshift__(self, other: int) -> Bits:
        return self.rshift(other)

    def __add__(self, other: Bits) -> Bits:
        return self.add_(other)

    def add_(self, other: Bits) -> Bits:
        return Bits(self.bitlist + other.bitlist)

    def __int__(self) -> int:
        return self.to_int()

    def __index__(self) -> int:
        return int(self)

    def __format__(self, format_spec) -> str:
        if format_spec == 'b':
            retstring = "0b"
            for i in range(len(self.bitlist)):
                if i % 8 == 0 and not i == 0:
                    retstring += "_"
                retstring += str(self.bitlist[i])
        elif format_spec == 'o':
            retstring = "0o"
            for i in range(0, len(self.bitlist), 3):
                if i % 8 == 0 and not i == 0:
                    retstring += "_"
                nibble = 0
                nibble_end_index = min(i + 3, len(self.bitlist))
                for bit in self.bitlist[i:nibble_end_index]:
                    nibble = (nibble << 1) | bit
                retstring += oct(nibble)[2:]
        elif format_spec == 'x':
            retstring = "0x"
            for i in range(0, len(self.bitlist), 4):
                if i % 8 == 0 and not i == 0:
                    retstring += "_"
                nibble = 0
                nibble_end_index = min(i + 4, len(self.bitlist))
                for bit in self.bitlist[i:nibble_end_index]:
                    nibble = (nibble << 1) | bit
                retstring += hex(nibble)[2:]
        return retstring

    def __bytes__(self) -> bytes:
        return self.to_bytes()

    def __Bits__(self) -> Bits:
        return self

    def append(self, value: int, inplace=True):
        value = int(value)
        if value not in [0, 1]:
            raise ValueError('Bits can only be 1 or 0')
        if inplace:
            self.bitlist.append(value)
            return self
        else:
            return Bits(self.bitlist + [value])

    def padleft(self, *, up_to_size: int, padvalue: int = 0, inplace=True):
        num_bits_to_add = up_to_size - len(self)
        if num_bits_to_add < 0:
            return self

        bits_to_add = [padvalue] * num_bits_to_add
        if inplace:
            self.bitlist = bits_to_add + self.bitlist
            return self
        else:
            return Bits(bits_to_add + self.bitlist)

    def padright(self, *, up_to_size: int, padvalue: int = 0, inplace=True):
        num_bits_to_add = up_to_size - len(self)
        if num_bits_to_add < 0:
            return self

        bits_to_add = [padvalue] * num_bits_to_add
        if inplace:
            self.bitlist = self.bitlist + bits_to_add
            return self
        else:
            return Bits(self.bitlist + bits_to_add)

    def pop(self, index: int = -1, inplace=True) -> int:
        if not inplace:
            bits_obj_to_act_on = Bits(self.bitlist, deep=True)
        else:
            bits_obj_to_act_on = self

        return bits_obj_to_act_on.bitlist.pop(index)

    def to_bytes(self, reverse_endianness=False) -> bytes:
        byte_arr = bytearray()
        for i in range(0, len(self.bitlist), 8):
            byte = 0
            byte_end_index = min(i + 8, len(self.bitlist))
            for bit in self.bitlist[i:byte_end_index]:
                byte = (byte << 1) | bit
            byte_arr.append(byte)

        if reverse_endianness:
            byte_arr = reversed(byte_arr)
        return bytes(byte_arr)

    @classmethod
    def from_bytes(cls, byte_arr: bytes, reverse_endianness=False):
        if reverse_endianness:
            byte_arr = reversed(byte_arr)

        return cls(byte_arr)

    @classmethod
    def from_string(cls, string: str):
        """
        Converts a string to a Bits object.
        The string must be all binary or of the form '0b', '0o', or '0x' followed by digits.
        """
        if "_" in string:
            string = string.replace("_", "")
        if string.startswith('0b'):
            binstring = string[2:]
        elif string.startswith('0o'):
            nonzeros = string[2:].lstrip('0')
            nonzeros_as_bin = bin(int(nonzeros, 8))[2:]
            leading_zeros_as_bin = '0' * (len(string[2:]) * 3 - len(nonzeros_as_bin))
            binstring = leading_zeros_as_bin + nonzeros_as_bin
        elif string.startswith('0x'):
            nonzeros = string[2:].lstrip('0')
            nonzeros_as_bin = bin(int(nonzeros, 16))[2:]
            leading_zeros_as_bin = '0' * (len(string[2:]) * 4 - len(nonzeros_as_bin))
            binstring = leading_zeros_as_bin + nonzeros_as_bin
        elif all(char in '01' for char in string):
            pass
        else:
            raise ValueError(
                f"Cannot convert {string} to Bits."
                f"Strings passed should be all binary or of the form '0b', '0o', or '0x' followed by digits."
            )

        # print("binstring: ", binstring)
        bin_list = [int(bit) for bit in binstring]
        # print(bin_list)
        return cls(bin_list)

    @classmethod
    def from_int(cls, integer: int, size=None):
        """
        Converts an integer to a Bits object. The size parameter determines the number of bits to use.
        If size is not provided, the number of bits required to represent the integer is used.
        """
        if size is None:
            size = twos_complement_bit_length(integer)
        if integer.bit_length() > size:
            raise ValueError(
                f"Cannot convert {integer} to Bits with size {size},"
                f" because it requires {integer.bit_length()} bits to represent.")

        bitlist = list()
        for index in range(size):
            bitlist.insert(0, (integer >> index) & 1)

        return cls(bitlist)

    def to_int(self, endianness='big', signed=True) -> int:
        """
        Converts a Bits object to an integer. It does this
            by casting the Bits to bytes, and then converting the bytes to an integer
            using the provided endianness and signedness.
        """
        copy = Bits(self.bitlist, deep=True)
        if signed and len(copy) > 0 and copy[0] == 1:
            next_multiple_of_8 = ceil(len(self.bitlist) / 8) * 8
            copy.padleft(up_to_size=next_multiple_of_8, padvalue=1, inplace=True)

        return int.from_bytes(copy.to_bytes(), byteorder=endianness, signed=signed)

    def to_bin(self) -> str:
        return self.str_(format_spec='b')

    def to_oct(self) -> str:
        return self.str_(format_spec='o')

    def to_hex(self) -> str:
        return self.str_(format_spec='x')

    def shrinkequals(self, other_bits: Bits) -> bool:
        """
        Returns True if both Bits objects are equal after removing leading zeros from the longer Bits object.
        """
        return self | other_bits == self & other_bits

    def __eq__(self, other: Bits):
        if not isinstance(other, Bits):
            return False
        return self.bitlist == other.bitlist

    def join(self, bitobjs: Iterable[Bits]) -> Bits:
        """
        Joins multiple Bits objects together with self in between each.
        """

        if len(self) == 0:
            fin_bit_list = [bit for bitobj in bitobjs for bit in bitobj]
            return Bits(fin_bit_list)
        else:
            ret_bits = Bits()
            for bitobj in list(bitobjs)[0:-1]:
                ret_bits = ret_bits + bitobj + self
            ret_bits = ret_bits + bitobjs[-1]
            return ret_bits




from __future__ import annotations
from abc import ABC, abstractmethod
import math
import struct
from typing import Any, Callable, Optional
from bytemaker.bits import Bits, BitsConstructorType
from bytemaker.utils import classproperty, ByteConvertible, is_instance_of_union


class CustomTypeTracker:
    """
    Class to track custom types created by the user.
    """
    bit_types = {}
    byte_types = {}
    str_types = {}
    uint_types = {}
    sint_types = {}
    float_types = {}


# Abstract Classes
class YType(ABC):
    """
    Abstract base class for all YType objects.

    YType objects are objects that can be converted to and from bits and bytes.

    Class Properties:
    -----------
    num_bits : int
        The number of bits in the YType object.
    num_bytes : int
        The number of bytes in the YType object.
    value_type : type
        The type of the value of the YType object.
    value : Any
        The value of the YType object.
    bytes : bytes
        The bytes representation of the YType object.
    bytearray : bytearray
        The bytearray representation of the YType object.
    bits : Bits
        The bitstring representation of the YType object.

    Required Abstract Methods:
    -----------------
    get_num_bits() -> int:
        Abstract class method that returns the number of bits in instances of the YType.

    get_value_type() -> type:
        Abstract class method that returns the type of the values of instances of the YType.


    to_bits(*args, **kwargs) -> Bits:
        Abstract instance method that converts the YType object to bitstring.Bits.

    from_bits(the_bits: Bits, *args, **kwargs):
        Abstract class method that converts bitstring.Bits to an instance of the YType.



    Concrete Methods:
    --------
    get_num_bytes() -> int:
       Class method that returns the number of bytes in instances of the YType.


    get_value() -> Any:
        Instance method that returns the value (internal) of the YType object.

    to_bytes(*args, **kwargs) -> bytes:
        Instance method that converts the YType object to bytes.

    from_bytes(the_bytes, *args, **kwargs):
        Class method that converts bytes to an instance of the YType class.

    __eq__(other) -> bool:
        Instance method that compares the YType object with another object.

    __Bits__() -> Bits:
        Instance method that converts the YType object to bitstring.Bits.
        Useful for Bits(theObject).

    __bytes__() -> bytes:
        Instance method that converts the YType object to bytes.
        Useful for bytes(theObject).

    __repr__() -> str:
        Instance method that returns the string representation of the object.
        Useful for repr(theObject), or if the class has not overridden __str__, str(thObject)
    """

    def __init__(self, value: bytes | bytearray | Bits | Any, test_creation=True, *args, **kwargs) -> None:
        """
        Initializes the YType object.

        Attempts the following:
        - If value is an instance of the valuetype of the YType subclass, sets the value directly.
        - If value is an instance of YType and the valuetype of both YTypes match, \
            sets the value to the value of the YType subclass.
        - If value is an instance of bitstring.Bits, calls from_bits(Bits(value)) from the YType subclass.
        - If value is an instance of bytes or bytearray, calls from_bytes(bytes(value)) from the YType subclass.
        - If value is an instance of BitsConstructorType, calls from_bits(Bits(value)) from the YType subclass.
        - If value is an instance of ByteConvertible, calls from_bytes(bytes(value)) from the YType subclass.
        - If value is castable to the valuetype of the YType subclass, sets the value to the casted value.
        - Otherwise, raises a TypeError.

        Args:
        - value: the value of the object.
        - test_creation: a boolean indicating whether to test the creation of the object.
            If true, calls to_bytes() to test if the object is properly convertible to bytes.
        """

        super().__init__()

        if self.value is None:
            if isinstance(value, self.value_type):
                self._value = value
            elif isinstance(value, YType) and self.value_type == value.value_type:
                self._value = value.value
            elif isinstance(value, Bits):
                self._value = self.from_bits(value).value
            elif isinstance(value, (bytes, bytearray)):
                self._value = self.from_bytes(value).value
            elif is_instance_of_union(value, BitsConstructorType):
                self._value = self.from_bits(Bits(value)).value
            elif isinstance(value, ByteConvertible):
                self._value = self.from_bytes(bytes(value)).value
            else:
                try:
                    newvalue = self.value_type(value)
                    self._value = newvalue
                except Exception as e:
                    raise TypeError(f"Expecting {self.value_type}, got {type(value)}"
                                    f"Exception details: {e}")

        if test_creation:
            self.validate_value(value)

    def validate_value(self, value):
        """
        Validates the value of the object. Called by __init__ if test_creation is True.
            By default, just tries to convert the object to bytes.
        """
        try:
            bytes(self)
        except Exception as e:
            raise type(e)(f"Object {value} is not properly convertible to bytes.")

    # Bits/bytes counting
    @classmethod
    @abstractmethod
    def get_num_bits(cls) -> int:
        """
        Returns the number of bits that represent the values of members of the class.
        """
        pass

    @classproperty
    def num_bits(cls) -> int:
        __doc__ = cls.get_num_bits.__doc__
        return cls.get_num_bits()

    @classmethod
    def get_num_bytes(cls) -> int:
        """
        Returns the number of bytes that represent the values of members of the class.
        """
        return cls.get_num_bits() / 8

    @classproperty
    def num_bytes(cls):
        __doc__ = cls.get_num_bytes.__doc__
        return cls.get_num_bytes()

    # Value type info
    @classmethod
    @abstractmethod
    def get_value_type(cls):
        """
        Returns the value type of the subclass.
        """
        pass

    @classproperty
    def value_type(cls):
        __doc__ = cls.get_value_type.__doc__
        return cls.get_value_type()

    # Value info
    def get_value(self):
        """
        Returns the value the instance of the subclass represents.
        """
        if hasattr(self, '_value'):
            return self._value
        else:
            return None

    @property
    def value(self):
        __doc__ = self.get_value.__doc__
        return self.get_value()

    # Conversions to/from bits and bytes
    @abstractmethod
    def to_bits(self, *args, **kwargs) -> Bits:
        """
        Returns the bitstring.Bits representation of the value the object represents
            in the number of bits specified by the class.
        """
        pass

    @classmethod
    @abstractmethod
    def from_bits(cls, the_bits: Bits, *args, **kwargs):
        """
        Returns an instance of the subclass representing the object represented by the_bits.
        """
        pass

    def __Bits__(self) -> Bits:
        """
        Returns the bitstring.Bits representation of the value the object represents
            Used by Bits(theObject)
        """
        return self.to_bits()

    def to_bytes(self, *args, **kwargs) -> bytes:
        """
        Returns the bytes representation of the value the object represents
            in the number of bytes specified by the class.
        """
        return bytes(self.to_bits(*args, **kwargs))

    @classmethod
    def from_bytes(cls, the_bytes, *args, **kwargs):
        """
        Returns an instance of the subclass representing the object represented by the_bytes.
        """
        return cls.from_bits(Bits(the_bytes), *args, **kwargs)

    def __bytes__(self) -> bytes:
        """
        Returns the bytes representation of the value the object represents
            Used by bytes(theObject)
        """
        return self.to_bytes()

    # Alternative bytes/bytearray/bits representations
    @property
    def bytes(self) -> bytes:
        __doc__ = self.to_bytes.__doc__
        return bytes(self)

    @property
    def bytearray(self) -> bytearray:
        """
        Returns a bytearray representation of the bytes representation of the value represented by the object.
        """
        return bytearray(bytes(self))

    @property
    def bits(self) -> Bits:
        """
        Returns a bitstring.Bits representation of the bytes representation of the value represented by the object.
        """
        return Bits(bytes(self))

    # Operators
    def __eq__(self, other) -> bool:
        """
        Returns whether the object is equal to another object.
            Here, this could mean the following, checked in order:
            - This object is a ytype and its value equals the other object
            - Both objects are ytypes and contain equivalent values.
                In the case of the values being floats, if both are nans, they are still considered equal
            - Both objects are implicitly bit-convertible through __Bits__() and have equivalent bit conversions
            - Both objects are implicitly byte-convertible through __bytes__() and have equivalent byte conversions
            - Both objects are the same object (default behavior of parent class)
        """

        if isinstance(other, self.value_type):
            return self.value == other
        elif isinstance(other, StructPackedYType):
            if self.value_type == other.value_type:
                both_floats = self.value_type == float and other.value_type == float
                both_nans = both_floats and math.isnan(self.value) and math.isnan(other.value)
                if not both_nans:
                    return self.value == other.value
        elif is_instance_of_union(other, BitsConstructorType):
            return Bits(self) == other.bits
        elif isinstance(other, ByteConvertible):
            return bytes(self) == bytes(other)

        else:
            return super().__eq__(other)

    # Other conversions
    def __repr__(self) -> str:
        """
        Returns the string representation of the object.

        Returns:
        - the string representation of the object.
        """
        classname = str(self.__class__).split('.')[-1].split("'")[0]
        return f"{classname}, value {str(self.value)}"


class StructPackedYType(YType):
    """
    Abstract base class for all YType objects that involve using struct for packing/unpacking.
    """

    def to_bits(self, *args, endianness='big', **kwargs):
        """
        Returns the bytes representation of the object.

        Args:
        - endianness: the endianness of the bytes representation.

        Returns:
        - the bytes representation of the object.
        """
        packing_format = self.get_packing_format(endianness)
        return Bits(struct.pack(packing_format, self.value, *args))

    @classmethod
    def from_bits(cls, the_bits, *args, endianness='big', **kwargs):
        """
        Returns the StructPackedYType object from the bytes representation.

        Args:
        - the_bytes: the bytes representation of the object.
        - endianness: the endianness of the bytes representation.

        Returns:
        - the StructPackedYType object from the bytes representation.
        """
        packing_format = cls.get_packing_format(endianness)
        return cls(struct.unpack(packing_format, bytes(the_bits), *args, **kwargs)[0])

    @abstractmethod
    def get_packing_format_letter(cls):
        """
        Returns the struct-packing format letter for the class.
        """
        pass

    @classproperty
    def packing_format_letter(cls):
        __doc__ = cls.get_packing_format_letter.__doc__
        return cls.get_packing_format_letter()

    @classmethod
    def get_packing_format(cls, endianness='big'):
        """
        Returns the packing format for the class.

        Args:
        - endianness: the endianness of the packing format. Defaults to little

        Returns:
        - the struct packing format for the subclass.
        """
        if endianness == 'little':
            return f"<{cls.packing_format_letter}"
        elif endianness == 'big':
            return f">{cls.packing_format_letter}"
        else:
            raise ValueError(f"Endianness must be either 'little' or 'big', not {endianness}")


# Concrete classes
# Bit Types
class BitYType(YType):
    """
    Abstract base class for all YType objects that represent bit values.
    """
    def to_bits(self, *args, **kwargs) -> Bits:
        default_bits = Bits(self, size=self.num_bits, *args, **kwargs)
        diff = self.num_bits - len(default_bits)
        if diff > 0:
            return default_bits + Bits([0] * diff)
        elif diff < 0:
            raise ValueError(
                f"Cannot convert {self} to Bits"
                f"because the number of bits in the object ({self.num_bits})"
                f"is greater than the number of bits in the object ({len(default_bits)})")
        else:
            return default_bits

    @classmethod
    def from_bits(cls, the_bits: Bits, *args, **kwargs):
        return cls(the_bits, *args, **kwargs)

    @classmethod
    def get_value_type(cls):
        return Bits


def BitsTypeFactory(size_in_bits: int):
    """
    Factory function for creating BitYType subclasses with the given number of bits.
        In the event there is an existing BitYType with the specified number of bits,
        retrieves it instead.
    Args:
    - size_in_bits: the number of bits the subclass represents.

    """
    init_type_name = f"Bit{size_in_bits}"
    cur_type_name = init_type_name
    cur_type_count = 0
    # Grab an existing type if this matches one. Ends with the "next" bit type name for bits with the given # of bits
    while cur_type_name in CustomTypeTracker.bit_types:
        if cur_type_name in CustomTypeTracker.bit_types[cur_type_name]:
            return CustomTypeTracker.bit_types[cur_type_name]
        cur_type_count += 1
        cur_type_name = f"{init_type_name}{cur_type_count}"

    class NewBitYType(Bits, BitYType):
        @classmethod
        def get_num_bits(cls) -> int:
            return size_in_bits

    NewBitYType.__name__ = cur_type_name
    CustomTypeTracker.bit_types[cur_type_name] = NewBitYType
    return NewBitYType


class Bit1(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 1
    pass


class Bit2(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 2
    pass


class Bit3(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 3
    pass


class Bit4(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 4
    pass


class Bit5(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 5
    pass


class Bit6(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 6
    pass


class Bit7(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 7
    pass


class Bit8(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 8
    pass


class Bit16(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 16
    pass


class Bit24(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 24
    pass


class Bit32(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 32
    pass


class Bit64(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 64
    pass


class Bit128(Bits, BitYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 128
    pass


# Byte Types
class ByteYType(YType, bytearray):
    """
    Abstract base class for all YType objects that represent raw byte values.
    """

    def to_bits(self, *args, **kwargs) -> Bits:
        default_bytes = self.value
        diff = math.ceil(self.num_bytes) - len(default_bytes)
        if diff > 0:
            return default_bytes + bytes(b'\x00' * diff)
        elif diff < 0:
            raise ValueError(
                f"Cannot convert {self} to Bits."
                f" The number of bits in the self.num_bits, ({self.num_bits})"
                f" is greater than the number of bits in the object ({len(default_bytes)})")
        else:
            return Bits(default_bytes)

    @classmethod
    def from_bits(cls, the_bits: Bits, *args, **kwargs):
        return cls(bytes(the_bits), *args, **kwargs)

    @classmethod
    def get_value_type(cls):
        return bytes


def BytesTypeFactory(size_in_bits: int):
    init_type_name = f"Byte{size_in_bits}"
    cur_type_name = init_type_name
    cur_type_count = 0
    # Grab an existing type if this matches one. Ends with the "next" byte type name for bytes with the given # of bits
    while cur_type_name in CustomTypeTracker.byte_types:
        if cur_type_name in CustomTypeTracker.byte_types:
            return CustomTypeTracker.byte_types[cur_type_name]
        cur_type_count += 1
        cur_type_name = f"{init_type_name}{cur_type_count}"

    class NewByteYType(ByteYType, bytearray):
        @classmethod
        def get_num_bits(cls) -> int:
            return size_in_bits

    NewByteYType.__name__ = cur_type_name
    CustomTypeTracker.byte_types[cur_type_name] = NewByteYType
    return NewByteYType


class Byte8(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 8
    pass


class Byte16(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 16
    pass


class Byte24(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 24
    pass


class Byte32(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 32
    pass


class Byte64(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 64
    pass


class Byte128(ByteYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 128
    pass


# Integer Types
class IntYType(YType):
    def __init__(self, value: int | bytes | bytearray | IntYType, *args, **kwargs):
        super().__init__(value, *args, **kwargs)

    @abstractmethod
    def get_is_signed(self):
        """
        Returns whether the subclass represents a signed (vs unsigned) integer.
        """
        pass

    @classproperty
    def is_signed(self):
        __doc__ = self.get_is_signed.__doc__
        pass

    @classmethod
    def get_value_type(cls):
        return int

    def __int__(self):
        return self.value


# UInts
class UInt(IntYType):
    @classmethod
    def get_is_signed(cls):
        return False


class UInt8(UInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 8

    @classmethod
    def get_packing_format_letter(cls):
        return 'B'


class UInt16(UInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 16

    @classmethod
    def get_packing_format_letter(cls):
        return 'H'


class UInt24(UInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 24

    @classmethod
    def get_packing_format_letter(cls):
        return 'I'


class UInt32(UInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 32

    @classmethod
    def get_packing_format_letter(cls):
        return 'I'


class UInt64(UInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 64

    @classmethod
    def get_packing_format_letter(cls):
        return 'Q'


# Signed ints
class SInt(IntYType):
    @classmethod
    def get_is_signed(cls):
        return True


class SInt8(SInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 8

    @classmethod
    def get_packing_format_letter(cls):
        return 'b'


class SInt16(SInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 16

    @classmethod
    def get_packing_format_letter(cls):
        return 'h'


class SInt24(SInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 24

    @classmethod
    def get_packing_format_letter(cls):
        return 'i'


class SInt32(SInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 32

    @classmethod
    def get_packing_format_letter(cls):
        return 'i'


class SInt64(SInt, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 64

    @classmethod
    def get_packing_format_letter(cls):
        return 'q'


# Floats
class FloatYType(YType):
    def __init__(self, value: float | bytes | bytearray | FloatYType, *args, **kwargs):
        super().__init__(value, *args, **kwargs)

    @classmethod
    def get_value_type(cls):
        return float

    def __float__(self):
        return self.value


class Float16(FloatYType, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 16

    @classmethod
    def get_packing_format_letter(cls):
        return 'e'


class Float32(FloatYType, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 32

    @classmethod
    def get_packing_format_letter(cls):
        return 'f'


class Float64(FloatYType, StructPackedYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 64

    @classmethod
    def get_packing_format_letter(cls):
        return 'd'


class StrYType(YType):
    def __init__(self, value: str | bytes | bytearray | StrYType, *args, **kwargs):
        super().__init__(value, *args, **kwargs)

    @property
    def value(self):
        return self._value

    @classmethod
    def get_value_type(cls):
        return str

    @classproperty
    def encoding_method(cls):
        return cls.get_encoding_method()

    @classproperty
    def decoding_method(cls):
        return cls.get_decoding_method()

    @classmethod
    @abstractmethod
    def get_encoding_method(cls):
        pass

    @classmethod
    @abstractmethod
    def get_decoding_method(cls):
        pass

    @classmethod
    def pack_string_same_size(cls, size_in_bits: int):
        packed_string = cls(size_in_bits)
        return packed_string

    def __str__(self):
        return self.value


class DefaultCodings:
    str_dec = lambda the_bits: bytes(the_bits).decode(encoding="UTF-8")
    str_enc = lambda the_str: Bits(the_str.encode(encoding="UTF-8"))


def StrTypeFactory(
        size_in_bits: int,
        encode_method: Callable[[str], Bits] = DefaultCodings.str_enc,
        decode_method: Callable[[Bits], str] = DefaultCodings.str_dec):
    init_type_name = f"Str{size_in_bits}"
    cur_type_name = init_type_name
    cur_type_count = 0
    # Grab an existing type if this matches one in name/encoding/decoding.
    # Ends with the "next" string type name for strings with the given number of bits
    while cur_type_name in CustomTypeTracker.str_types:
        if (
            CustomTypeTracker.str_types[cur_type_name].encode_method == encode_method and
            CustomTypeTracker.str_types[cur_type_name].decode_method == decode_method
        ):
            return CustomTypeTracker.str_types[cur_type_name]
        cur_type_count += 1
        cur_type_name = f"{init_type_name}Coding{cur_type_count}"

    class NewStrYType(StrYType):
        def __init__(self, value: str | Bits | bytes | bytearray | StrYType, *args, **kwargs):
            if isinstance(value, str):
                self._value = value
            elif isinstance(value, bytes) or isinstance(value, bytearray) or isinstance(value, Bits):
                self._value = decode_method(value, *args, **kwargs)

            super().__init__(value, *args, **kwargs)

        @classmethod
        def get_num_bits(cls) -> int:
            return size_in_bits

        def to_bits(self, *args, **kwargs) -> Bits:
            str_len = len(encode_method(self.value))
            diff = self.num_bits - str_len
            default_bits = Bits(encode_method(self.value), *args, **kwargs)

            # print("before", len(default_bits))
            if diff > 0:
                default_bits = default_bits + [0] * diff
            # print(len(default_bits))
            return default_bits

        def get_encoding_method(cls):
            return encode_method

        def get_decoding_method(cls):
            return decode_method

        @classmethod
        def from_bits(cls, the_bits, *args, **kwargs):
            return cls(decode_method(the_bits, *args, **kwargs))

        def __repr__(self):
            str_len = len(encode_method(self.value))
            diff = self.num_bits - str_len
            addendum_bytes = ""
            addendum_bits = ""

            if diff > 0:
                initial_counter = str_len
                addendum_bytes = "\\x"
                while initial_counter < self.num_bits:
                    if initial_counter % 32 == 0:
                        addendum_bytes += "_"
                    addendum_bytes += "0"
                    initial_counter += 4

                if initial_counter > self.num_bits:
                    initial_counter -= 4
                    addendum_bits = "\\b"
                    while initial_counter < self.num_bits:
                        if initial_counter % 8 == 0:
                            addendum_bits += "_"
                        addendum_bits += "0"
                        initial_counter += 1

                return self.value + addendum_bytes + addendum_bits
            else:
                return self.value

        def __str__(self):
            return self.value

    NewStrYType.__name__ = cur_type_name
    NewStrYType.encode_method = encode_method
    NewStrYType.decode_method = decode_method
    CustomTypeTracker.str_types[cur_type_name] = NewStrYType
    return NewStrYType


class Str8(StrTypeFactory(8), StrYType):
    @classmethod
    def get_num_bits(cls) -> int:
        return 8


CustomTypeTracker.str_types = {}
CustomTypeTracker.bit_types = {
    Bit1.__name__: Bit1,
    Bit2.__name__: Bit2,
    Bit3.__name__: Bit3,
    Bit4.__name__: Bit4,
    Bit5.__name__: Bit5,
    Bit6.__name__: Bit6,
    Bit7.__name__: Bit7,
    Bit8.__name__: Bit8,
    Bit16.__name__: Bit16,
    Bit24.__name__: Bit24,
    Bit32.__name__: Bit32,
    Bit64.__name__: Bit64,
    Bit128.__name__: Bit128
}
CustomTypeTracker.byte_types = {
    Byte8.__name__: Byte8,
    Byte16.__name__: Byte16,
    Byte24.__name__: Byte24,
    Byte32.__name__: Byte32,
    Byte64.__name__: Byte64,
    Byte128.__name__: Byte128
}
CustomTypeTracker.uint_types = {
    UInt8.__name__: UInt8,
    UInt16.__name__: UInt16,
    UInt24.__name__: UInt24,
    UInt32.__name__: UInt32,
    UInt64.__name__: UInt64
}
CustomTypeTracker.sint_types = {
    SInt8.__name__: SInt8,
    SInt16.__name__: SInt16,
    SInt24.__name__: SInt24,
    SInt32.__name__: SInt32,
    SInt64.__name__: SInt64
}
CustomTypeTracker.float_types = {
    Float16.__name__: Float16,
    Float32.__name__: Float32,
    Float64.__name__: Float64
}


def ytype_to_bytes(unit: YType, num_bytes: Optional[int] = None, reverse_endianness: bool = False) -> bytes:
    if isinstance(unit, YType):
        return unit.to_bytes(num_bytes=num_bytes, reverse_endianness=reverse_endianness)
    else:
        raise TypeError(f"Expecting YType, got {type(unit)}")


def bytes_to_ytype(unitbytes: bytes, unittype: YType, reverse_endianness: bool = False) -> YType:
    if reverse_endianness:
        unitbytes = unitbytes[::-1]

    if issubclass(unittype, YType):
        unit = unittype.from_bytes(unitbytes)
    else:
        raise TypeError(f"Expecting YType, got {unittype}")

    return unit


if __name__ == "__main__":
    the_bytes = Bits(StrTypeFactory(10 * 8)("fish"))
    print(the_bytes)
    print(bytes_to_ytype(the_bytes, StrTypeFactory(20), reverse_endianness=False))

    print(Str8("f"))



import dataclasses
import typing
import warnings
from typing import Any
from math import log2, ceil

#  General Python functionality


class classproperty(property):
    """
    Class property decorator.
    """
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


def is_instance_of_union(obj, union_type: type):
    """
    Determines if an object is an instance of a union type (to support Python versions <3.10).

    Args:
        obj (object): The object to check
        union_type (type): The union type to check against

    Returns:
        bool: Whether the object is an instance of the union type
    """

    if typing.get_origin(union_type) is not typing.Union:
        warnings.warn(
            f"Checking for instance of a union type with the non-union type {union_type}"
            f" with origin {typing.get_origin(union_type)}. Falling back on isinstance")
        return isinstance(obj, union_type)
    return any(isinstance(obj, union_component_type) for union_component_type in typing.get_args(union_type))


def is_subclass_of_union(obj_type: type, union_type: type):
    """
    Determines if an object is a subclass of a union type (to support Python versions <3.10).

    Args:
        obj_type (type): The object type to check
        union_type (type): The union type to check against

    Returns:
        bool: Whether the object is a subclass of the union type
    """

    if typing.get_origin(union_type) is not typing.Union:
        warnings.warn(f"Checking for subclass of a union type with the non-union type {union_type}."
                      f"Falling back on issubclass.")
        return issubclass(obj_type, union_type)
    return any(issubclass(obj_type, union_type_part) for union_type_part in typing.get_args(union_type))

# General operations


# Byte-related operations

class ByteConvertibleMeta(type):
    """
    This is used to create IsByteConvertible, a type to allow checking \
        whether an object or instances of a class can be converted to a bytes object using isinstance or issubclass.
    """
    def __instancecheck__(self, __instance: Any) -> bool:
        try:
            bytes(__instance)
            return True
        except Exception:
            return False

    def __subclasscheck__(self, __subclass: type) -> bool:
        return (
            hasattr(__subclass, '__bytes__') or
            is_subclass_of_union(__subclass, typing.Union[bytes, bytearray, memoryview]) or
            (hasattr(__subclass, '__getitem__') and hasattr(__subclass, 'format') and
                hasattr(__subclass, 'shape') and hasattr(__subclass, 'strides'))
        )


class ByteConvertible(metaclass=ByteConvertibleMeta):
    pass


# Aggregate type to bytes
class DataClassTypeMeta(type):
    def __instancecheck__(cls, instance):
        return dataclasses.is_dataclass(instance)

    def __subclasscheck__(cls, subclass):
        return dataclasses.is_dataclass(subclass)


class DataClassType(metaclass=DataClassTypeMeta):
    pass


def twos_complement_bit_length(n: int):
    """
    Determines the number of bits required to represent a signed integer in two's complement.

    Args:
        n (int): The (signed) integer for which to determine the number of bits required

    Returns:
        int: The number of bits required to represent the integer in two's-complement notation
    """
    if n == 0:
        return 1  # Technically can represent 0 with 0 bits in two's complement, but this is not useful

    is_greq_than_zero = n >= 0
    abs_val = abs(n)
    is_power_of_two = (abs_val & (abs_val - 1)) == 0

    if is_greq_than_zero or not is_power_of_two:
        return ceil(log2(abs_val + 1)) + 1  # Account for extra at negative extreme
    else:
        return int(log2(abs_val)) + 1


def twos_complement(number, n_bits=32):
    """
    Convert an integer to its two's complement representation.

    :param number: The integer to convert.
    :param bits: The bit width for the two's complement representation.
    :return: A string representing the two's complement of the number.
    """
    if number < 0:
        number = (1 << n_bits) + number
    format_string = '{:0' + str(n_bits) + 'b}'
    return format_string.format(number)
